from .naming import frozendict


class TypeCheck(object):
  """
    Encapsulate the results of a type check pass.
  """
  class Error(Exception):
    pass

  @staticmethod
  def success():
    return TypeCheck(True, "")

  @staticmethod
  def failure(msg):
    return TypeCheck(False, msg)

  def __init__(self, success, message):
    self._success = success
    self._message = message

  def message(self):
    return self._message

  def ok(self):
    return self._success

  def __repr__(self):
    if self.ok():
      return 'TypeCheck(OK)'
    else:
      return 'TypeCheck(FAILED): %s' % self._message


class TypeFactoryType(type):
  _TYPE_FACTORIES = {}

  def __new__(mcs, name, parents, attributes):
    """Args:
       mcs(metaclass): the class object to create an instance of. Since this is actually
           creating an instance of a type factory class, it's really a metaclass.
       name (str): the name of the type to create.
       parents (list(class)): the superclasses.
       attributes (map(string, value)):
    """
    if 'PROVIDES' not in attributes:
      return type.__new__(mcs, name, parents, attributes)
    else:
      provides = attributes['PROVIDES']
      new_type = type.__new__(mcs, name, parents, attributes)
      TypeFactoryType._TYPE_FACTORIES[provides] = new_type
      return new_type


TypeFactoryClass = TypeFactoryType('TypeFactoryClass', (object,), {})
class TypeFactory(TypeFactoryClass):
  @staticmethod
  def get_factory(type_name):
    assert type_name in TypeFactoryType._TYPE_FACTORIES, (
      'Unknown type: %s, Existing factories: %s' % (
        type_name, TypeFactoryType._TYPE_FACTORIES.keys()))
    return TypeFactoryType._TYPE_FACTORIES[type_name]

  @staticmethod
  def create(type_dict, *type_parameters):
    """
      Implemented by the TypeFactory to produce a new type.

      Should return:
        reified type
        (with usable type.__name__)
    """
    raise NotImplementedError("create unimplemented for: %s" % repr(type_parameters))

  @staticmethod
  def new(type_dict, type_factory, *type_parameters):
    """
      Create a fully reified type from a type schema.
    """
    type_tuple = (type_factory,) + type_parameters
    if type_tuple not in type_dict:
      factory = TypeFactory.get_factory(type_factory)
      reified_type = factory.create(type_dict, *type_parameters)
      type_dict[type_tuple] = reified_type
    return type_dict[type_tuple]

  @staticmethod
  def wrapper(factory):
    assert issubclass(factory, TypeFactory)
    def wrapper_function(*type_parameters):
      return TypeFactory.new({}, factory.PROVIDES, *tuple(
        [typ.serialize_type() for typ in type_parameters]))
    return wrapper_function

  @staticmethod
  def load(type_tuple, into=None):
    """
      Determine all types touched by loading the type and deposit them into
      the particular namespace.
    """
    type_dict = {}
    TypeFactory.new(type_dict, *type_tuple)
    deposit = into if (into is not None and isinstance(into, dict)) else {}
    for reified_type in type_dict.values():
      deposit[reified_type.__name__] = reified_type
    return deposit

  @staticmethod
  def load_json(json_list, into=None):
    """
      Determine all types touched by loading the type and deposit them into
      the particular namespace.
    """
    def l2t(obj):
      if isinstance(obj, list):
        return tuple(l2t(L) for L in obj)
      elif isinstance(obj, dict):
        return frozendict(obj)
      else:
        return obj
    return TypeFactory.load(l2t(json_list), into=into)

  @staticmethod
  def load_file(filename, into=None):
    import json
    with open(filename) as fp:
      return TypeFactory.load_json(json.load(fp), into=into)


class TypeMetaclass(type):
  def __instancecheck__(cls, other):
    if not hasattr(other, 'type_parameters'):
      return False
    if not hasattr(other, '__class__'):
      return False
    if cls.__name__ != other.__class__.__name__:
      return False
    return cls.type_factory() == other.type_factory() and (
      cls.type_parameters() == other.type_parameters())

  def __new__(mcls, name, parents, attributes):
    """Creates a new Type object (an instance of TypeMetaclass).
    Args:
        name (str): the name of the new type.
        parents (list(str)): a list of superclasses.
        attributes: (???): a map from name to value for "parameters" for defining
           the new type. 
    """
    return type.__new__(mcls, name, parents, attributes)


class Type(object):
  @classmethod
  def type_factory(cls):
    """ Return the name of the factory that produced this class. """
    raise NotImplementedError

  @classmethod
  def type_parameters(cls):
    """ Return the type parameters used to produce this class. """
    raise NotImplementedError

  @classmethod
  def serialize_type(cls):
    return (cls.type_factory(),) + cls.type_parameters()

  @classmethod
  def dump(cls, fp):
    import json
    json.dump(cls.serialize_type(), fp)

  def check(self):
    """
      Returns a TypeCheck object explaining whether or not a particular
      instance of this object typechecks.
    """
    raise NotImplementedError



import copy
from collections import Iterable, Mapping, Sequence
from inspect import isclass

from .base import Object
from .compatibility import Compatibility
from .naming import Namable, frozendict
from .typing import Type, TypeCheck, TypeFactory, TypeMetaclass


class ListFactory(TypeFactory):
  PROVIDES = 'List'

  @staticmethod
  def create(type_dict, *type_parameters):
    """
      Construct a List containing type 'klazz'.
    """
    assert len(type_parameters) == 1
    klazz = TypeFactory.new(type_dict, *type_parameters[0])
    assert isclass(klazz)
    assert issubclass(klazz, Object)
    return TypeMetaclass('%sList' % klazz.__name__, (ListContainer,), {'TYPE': klazz})


class ListContainer(Object, Namable, Type):
  """
    The List container type.  This is the base class for all user-generated
    List types.  It won't function as-is, since it requires cls.TYPE to be
    set to the contained type.  If you want a concrete List type, see the
    List() function.
  """
  __slots__ = ('_values',)

  def __init__(self, vals):
    self._values = self._coerce_values(copy.copy(vals))
    super(ListContainer, self).__init__()

  def get(self):
    return tuple(v.get() for v in self._values)

  def dup(self):
    return self.__class__(self._values)

  def __hash__(self):
    return hash(self.get())

  def __repr__(self):
    si, _ = self.interpolate()
    return '%s(%s)' % (self.__class__.__name__,
      ', '.join(str(v) if Compatibility.PY3 else unicode(v) for v in si._values))

  def __iter__(self):
    si, _ = self.interpolate()
    return iter(si._values)

  def __getitem__(self, index_or_slice):
    si, _ = self.interpolate()
    return si._values[index_or_slice]

  def __contains__(self, item):
    si, _ = self.interpolate()
    if isinstance(item, self.TYPE):
      return item in si._values
    else:
      return item in si.get()

  def __eq__(self, other):
    if not isinstance(other, ListContainer): return False
    if self.TYPE.serialize_type() != other.TYPE.serialize_type(): return False
    si, _ = self.interpolate()
    oi, _ = other.interpolate()
    return si._values == oi._values

  @staticmethod
  def isiterable(values):
    return isinstance(values, Sequence) and not isinstance(values, Compatibility.stringy)

  def _coerce_values(self, values):
    if not ListContainer.isiterable(values):
      raise ValueError("ListContainer expects an iterable, got %s" % repr(values))
    def coerced(value):
      return value if isinstance(value, self.TYPE) else self.TYPE(value)
    return tuple([coerced(v) for v in values])

  def check(self):
    assert ListContainer.isiterable(self._values)
    scopes = self.scopes()
    for element in self._values:
      assert isinstance(element, self.TYPE)
      typecheck = element.in_scope(*scopes).check()
      if not typecheck.ok():
        return TypeCheck.failure("Element in %s failed check: %s" % (self.__class__.__name__,
          typecheck.message()))
    return TypeCheck.success()

  def interpolate(self):
    unbound = set()
    interpolated = []
    scopes = self.scopes()
    for element in self._values:
      einterp, eunbound = element.in_scope(*scopes).interpolate()
      interpolated.append(einterp)
      unbound.update(eunbound)
    return self.__class__(interpolated), list(unbound)

  def find(self, ref):
    if not ref.is_index():
      raise Namable.NamingError(self, ref)
    try:
      intvalue = int(ref.action().value)
    except ValueError:
      raise Namable.NamingError(self, ref)
    if len(self._values) <= intvalue:
      raise Namable.NotFound(self, ref)
    else:
      namable = self._values[intvalue]
      if ref.rest().is_empty():
        return namable.in_scope(*self.scopes())
      else:
        if not isinstance(namable, Namable):
          raise Namable.Unnamable(namable)
        else:
          return namable.in_scope(*self.scopes()).find(ref.rest())

  @classmethod
  def type_factory(cls):
    return 'List'

  @classmethod
  def type_parameters(cls):
    return (cls.TYPE.serialize_type(),)

List = TypeFactory.wrapper(ListFactory)


class MapFactory(TypeFactory):
  PROVIDES = 'Map'

  @staticmethod
  def create(type_dict, *type_parameters):
    assert len(type_parameters) == 2, 'Type parameters: %s' % repr(type_parameters)
    key_klazz, value_klazz = type_parameters
    key_klazz, value_klazz = (TypeFactory.new(type_dict, *key_klazz),
                              TypeFactory.new(type_dict, *value_klazz))
    assert isclass(key_klazz) and isclass(value_klazz)
    assert issubclass(key_klazz, Object) and issubclass(value_klazz, Object)
    return TypeMetaclass('%s%sMap' % (key_klazz.__name__, value_klazz.__name__), (MapContainer,),
      {'KEYTYPE': key_klazz, 'VALUETYPE': value_klazz})


# TODO(wickman) Technically it's possible to do the following:
#
# >>> my_map = Map(Boolean,Integer)((True,2), (False,3), (False, 2))
# >>> my_map
# BooleanIntegerMap(True => 2, False => 3, False => 2)
# >>> my_map.get()
# frozendict({False: 2, True: 2})
# >>> my_map[True]
# Integer(2)
# >>> my_map.get()[True]
# 2
# we should filter tuples for uniqueness.
class MapContainer(Object, Namable, Type):
  """
    The Map container type.  This is the base class for all user-generated
    Map types.  It won't function as-is, since it requires cls.KEYTYPE and
    cls.VALUETYPE to be set to the appropriate types.  If you want a
    concrete Map type, see the Map() function.

    __init__(dict) => translates to list of tuples & sanity checks
    __init__(tuple) => sanity checks
  """
  __slots__ = ('_map',)

  def __init__(self, *args):
    """
      Construct a map.

      Input:
        sequence of tuples _or_ a dictionary
    """
    if len(args) == 1 and isinstance(args[0], Mapping):
      self._map = self._coerce_map(copy.copy(args[0]))
    elif all(isinstance(arg, Iterable) and len(arg) == 2 for arg in args):
      self._map = self._coerce_tuple(args)
    else:
      raise ValueError("Unexpected input to MapContainer: %s" % repr(args))
    super(MapContainer, self).__init__()

  def get(self):
    return frozendict((k.get(), v.get()) for (k, v) in self._map)

  def _coerce_wrapper(self, key, value):
    coerced_key = key if isinstance(key, self.KEYTYPE) else self.KEYTYPE(key)
    coerced_value = value if isinstance(value, self.VALUETYPE) else self.VALUETYPE(value)
    return (coerced_key, coerced_value)

  def _coerce_map(self, input_map):
    return tuple(self._coerce_wrapper(key, value) for key, value in input_map.items())

  def _coerce_tuple(self, input_tuple):
    return tuple(self._coerce_wrapper(key, value) for key, value in input_tuple)

  def __hash__(self):
    return hash(self.get())

  def __iter__(self):
    si, _ = self.interpolate()
    return (t[0] for t in si._map)

  def __getitem__(self, key):
    if not isinstance(key, self.KEYTYPE):
      try:
        key = self.KEYTYPE(key)
      except ValueError:
        raise KeyError("%s is not coercable to %s" % self.KEYTYPE.__name__)
    # TODO(wickman) The performance of this should be improved.
    si, _ = self.interpolate()
    for tup in si._map:
      if key == tup[0]:
        return tup[1]
    raise KeyError("%s not found" % key)

  def __contains__(self, item):
    try:
      self[item]
      return True
    except KeyError:
      return False

  def dup(self):
    return self.__class__(*self._map)

  def __repr__(self):
    si, _ = self.interpolate()
    return '%s(%s)' % (self.__class__.__name__,
      ', '.join('%s => %s' % (key, val) for key, val in si._map))

  def __eq__(self, other):
    if not isinstance(other, MapContainer): return False
    if self.KEYTYPE.serialize_type() != other.KEYTYPE.serialize_type(): return False
    if self.VALUETYPE.serialize_type() != other.VALUETYPE.serialize_type(): return False
    si, _ = self.interpolate()
    oi, _ = other.interpolate()
    return si._map == oi._map

  def check(self):
    assert isinstance(self._map, tuple)
    scopes = self.scopes()
    for key, value in self._map:
      assert isinstance(key, self.KEYTYPE)
      assert isinstance(value, self.VALUETYPE)
      keycheck = key.in_scope(*scopes).check()
      valuecheck = value.in_scope(*scopes).check()
      if not keycheck.ok():
        return TypeCheck.failure("%s key %s failed check: %s" % (self.__class__.__name__,
          key, keycheck.message()))
      if not valuecheck.ok():
        return TypeCheck.failure("%s[%s] value %s failed check: %s" % (self.__class__.__name__,
          key, value, valuecheck.message()))
    return TypeCheck.success()

  def interpolate(self):
    unbound = set()
    interpolated = []
    scopes = self.scopes()
    for key, value in self._map:
      kinterp, kunbound = key.in_scope(*scopes).interpolate()
      vinterp, vunbound = value.in_scope(*scopes).interpolate()
      unbound.update(kunbound)
      unbound.update(vunbound)
      interpolated.append((kinterp, vinterp))
    return self.__class__(*interpolated), list(unbound)

  def find(self, ref):
    if not ref.is_index():
      raise Namable.NamingError(self, ref)
    kvalue = self.KEYTYPE(ref.action().value)
    scopes = self.scopes()
    for key, namable in self._map:
      if kvalue == key:
        if ref.rest().is_empty():
          return namable.in_scope(*scopes)
        else:
          if not isinstance(namable, Namable):
            raise Namable.Unnamable(namable)
          else:
            return namable.in_scope(*scopes).find(ref.rest())
    raise Namable.NotFound(self, ref)

  @classmethod
  def type_factory(cls):
    return 'Map'

  @classmethod
  def type_parameters(cls):
    return (cls.KEYTYPE.serialize_type(), cls.VALUETYPE.serialize_type())


Map = TypeFactory.wrapper(MapFactory)



# Choice types: types that can take one of a group of selected types.


from .base import Object
from .compatibility import Compatibility
from .typing import Type, TypeCheck, TypeFactory, TypeMetaclass


class ChoiceFactory(TypeFactory):
  """A Pystachio type representing a value which can be one of several
  different types.
  For example, a field which could be either an integer, or an integer
  expression (where IntegerExpression is a struct type) could be written
  Choice("IntOrExpr", (Integer, IntegerExpression))
  """
  PROVIDES = 'Choice'

  @staticmethod
  def create(type_dict, *type_parameters):
    """
    type_parameters should be:
      (name, (alternative1, alternative2, ...))
    where name is a string, and the alternatives are all valid serialized
    types.
    """
    assert len(type_parameters) == 2
    name = type_parameters[0]
    alternatives = type_parameters[1]
    assert isinstance(name, Compatibility.stringy)
    assert isinstance(alternatives, (list, tuple))
    choice_types = []
    for c in alternatives:
      choice_types.append(TypeFactory.new(type_dict, *c))
    return TypeMetaclass(str(name), (ChoiceContainer,), {'CHOICES': choice_types})


class ChoiceContainer(Object, Type):
  """The inner implementation of a choice type value.

  This just stores a value, and then tries to coerce it into one of the alternatives when
  it's checked or interpolated.
  """
  __slots__ = ('_value',)

  def __init__(self, val):
    super(ChoiceContainer, self).__init__()
    self._value = val

  def get(self):
    return self.unwrap().get()

  def unwrap(self):
    """Get the Pystachio value that's wrapped in this choice."""
    return self.interpolate()[0]

  def dup(self):
    return self.__class__(self._value)

  def __hash__(self):
    return hash(self.get())

  def __unicode__(self):
    return unicode(self.unwrap())

  def __str__(self):
    return str(self.unwrap())

  def __repr__(self):
    return '%s(%s)' % (self.__class__.__name__,
                       repr(self._value))

  def __eq__(self, other):
    if not isinstance(other, ChoiceContainer):
      return False
    if len(self.CHOICES) != len(other.CHOICES):
      return False
    for myalt, otheralt in zip(self.CHOICES, other.CHOICES):
      if myalt.serialize_type() != otheralt.serialize_type():
        return False
    si, _ = self.interpolate()
    oi, _ = other.interpolate()
    return si == oi

  def _unwrap(self, ret_fun, err_fun):
    """Iterate over the options in the choice type, and try to perform some
    action on them. If the action fails (returns None or raises either CoercionError
    or ValueError), then it goes on to the next type.
    Args:
       ret_fun: a function that takes a wrapped option value, and either returns a successful
           return value or fails.
       err_fun: a function that takes the unwrapped value of this choice, and generates
           an appropriate error.
    Returns: the return value from a successful invocation of ret_fun on one of the
       type options. If no invocation fails, then returns the value of invoking err_fun.
    """
    for opt in self.CHOICES:
      if isinstance(self._value, opt):
        return ret_fun(self._value)
      else:
        try:
          o = opt(self._value)
          ret = ret_fun(o)
          if ret:
            return ret
        except (self.CoercionError, ValueError):
          pass
    return err_fun(self._value)

  def check(self):
    # Try each of the options in sequence:
    # There are three cases for matching depending on the value:
    # (1) It's a pystachio value, and its type is the type alternative. Then typecheck
    #  succeeds.
    # (2) It's a pystachio value, but its type is not the current alternative. Then the
    #  typecheck proceeds to the next alternative.
    # (3) It's not a pystachio value. Then we try to coerce it to the type alternative.
    #  If it succeeds, then the typecheck succeeds. Otherwise, it proceeds to the next
    #  type alternative.
    # If none of the type alternatives succeed, then the check fails. match
    def _check(v):
      tc = v.in_scope(*self.scopes()).check()
      if tc.ok():
        return tc

    def _err(v):
      return TypeCheck.failure(
        "%s typecheck failed: value %s did not match any of its alternatives" %
        (self.__class__.__name__, v))

    return self._unwrap(_check, _err)

  def interpolate(self):
    def _inter(v):
      return v.in_scope(*self.scopes()).interpolate()

    def _err(v):
      raise self.CoercionError(self._value, self.__class__)

    return self._unwrap(_inter, _err)

  @classmethod
  def type_factory(cls):
    return 'Choice'

  @classmethod
  def type_parameters(cls):
    tup = tuple(t.serialize_type() for t in cls.CHOICES)
    return (cls.__name__, tup)

  @classmethod
  def serialize_type(cls):
    return (cls.type_factory(),) + cls.type_parameters()


def Choice(*args):
  """Helper function for creating new choice types.
  This can be called either as:
     Choice(Name, [Type1, Type2, ...])
  or:
     Choice([Type1, Type2, ...])
  In the latter case, the name of the new type will be autogenerated, and will
  look like "Choice_Type1_Type2".
  """
  if len(args) == 2:
    name, alternatives = args
  else:
    name = "Choice_" + "_".join(a.__name__ for a in args[0])
    alternatives = args[0]
  assert isinstance(name, Compatibility.stringy)
  assert all(issubclass(t, Type) for t in alternatives)
  return TypeFactory.new({}, ChoiceFactory.PROVIDES, name,
                         tuple(t.serialize_type() for t in alternatives))



import copy
import re

from .compatibility import Compatibility
from .naming import Ref

try:
  from itertools import izip_longest as zipl
except ImportError:
  from itertools import zip_longest as zipl



class Any(object):
  pass


class Matcher(object):
  """
    Matcher of Ref patterns.
  """

  __slots__ = ('_components', '_resolver')

  class Error(Exception): pass
  class NoMatch(Error): pass

  @classmethod
  def escape(cls, pattern):
    if not pattern.endswith('$'):
      pattern = pattern + '$'
    if not pattern.startswith('^'):
      pattern = '^' + pattern
    return re.compile(pattern)

  def __init__(self, root=None):
    if root is None:
      self._components = []
      return
    if not isinstance(root, Compatibility.stringy) and root is not Any:
      raise ValueError('Invalid root match value: %s' % root)
    self._components = [Ref.Dereference(self.escape('.*' if root is Any else root))]

  def __extend(self, value):
    new_match = Matcher()
    new_match._components = copy.copy(self._components) + [value]
    return new_match

  def __getattr__(self, pattern):
    if pattern == '_':
      return lambda pattern: self.__extend(Ref.Dereference(self.escape(pattern)))
    elif pattern == 'Any':
      return self.__extend(Ref.Dereference(self.escape('.*')))
    else:
      return self.__extend(Ref.Dereference(self.escape(pattern)))

  def __getitem__(self, pattern):
    if pattern is Any:
      return self.__extend(Ref.Index(self.escape('.*')))
    else:
      return self.__extend(Ref.Index(self.escape(pattern)))

  def __repr__(self):
    return 'Match(%s)' % '+'.join(map(str, self._components))

  def match(self, pystachio_object):
    _, refs = pystachio_object.interpolate()
    for ref in refs:
      args = []
      zips = list(zipl(self._components, ref.components()))
      for pattern, component in zips[:len(self._components)]:
        if pattern.__class__ != component.__class__ or not pattern.value.match(component.value):
          break
        args.append(component.value)
      else:
        yield tuple(args)

  def __translate(self, match_tuple):
    components = []
    for component, match in zip(self._components, match_tuple):
      components.append(component.__class__(match))
    return Ref(components)

  def apply(self, binder, pystachio_object):
    if not callable(binder):
      raise TypeError('binder must be a callable')
    for match in self.match(pystachio_object):
      pystachio_object = pystachio_object.bind({self.__translate(match): binder(*match)})
    return pystachio_object



from numbers import Integral, Real
from sys import version_info


class Compatibility(object):
  """2.x + 3.x compatibility"""
  stringy = (str,)
  if version_info[0] == 2:
    stringy = (str, unicode)
  integer = (Integral,)
  real = (Real,)
  numeric = integer + real
  PY2 = version_info[0] == 2
  PY3 = version_info[0] == 3
  if PY3:
    @staticmethod
    def exec_function(ast, globals_map):
      locals_map = globals_map
      exec(ast, globals_map, locals_map)
      return locals_map
  else:
    eval(compile(
"""
@staticmethod
def exec_function(ast, globals_map):
  locals_map = globals_map
  exec ast in globals_map, locals_map
  return locals_map
""", "<exec_function>", "exec"))



import os
from functools import reduce

from .compatibility import Compatibility

try:
  import pkg_resources
except ImportError:
  pkg_resources = None


def relativize(from_path, include_path):
  return os.path.join(os.path.dirname(from_path), include_path)


class ConfigContext(object):
  ROOT = ''

  # Make JSON-friendly keys -- since JSON can't encode tuples as dictionary keys.
  @classmethod
  def key(cls, from_path, include_string):
    return '\0'.join([from_path, include_string])

  @classmethod
  def from_key(cls, key):
    return key.split('\0')

  def __init__(self, environment, loadables):
    self.environment = environment
    self.loadables = loadables

  def compile(self, from_path, include_string, data):
    self.loadables[self.key(from_path, include_string)] = data
    Compatibility.exec_function(compile(data, include_string, 'exec'), self.environment)


class ConfigExecutor(object):
  ROOT = ConfigContext.ROOT

  @classmethod
  def matches(cls, loadable):
    return False

  @classmethod
  def get(cls, loadable):
    """return the include function and the root include object."""
    raise NotImplementedError


class FileExecutor(ConfigExecutor):
  @classmethod
  def matches(cls, loadable):
    return isinstance(loadable, Compatibility.stringy) and os.path.isfile(loadable)

  @classmethod
  def compile_into(cls, context, from_path, config_file):
    actual_file = relativize(from_path, config_file)
    with open(actual_file) as fp:
      context.compile(from_path, config_file, fp.read())

  @classmethod
  def get(cls, loadable):
    deposit_stack = [cls.ROOT]
    def ast_executor(config_file, context):
      from_path = deposit_stack[-1]
      actual_file = relativize(from_path, config_file)
      deposit_stack.append(actual_file)
      cls.compile_into(context, from_path, config_file)
      deposit_stack.pop()
    return ast_executor, loadable


class ResourceExecutor(FileExecutor):
  @classmethod
  def resource_exists(cls, loadable):
    if pkg_resources is None:
      return False
    module_base, module_file = os.path.split(loadable)
    module_base = module_base.replace(os.sep, '.')
    try:
      return pkg_resources.resource_exists(module_base, module_file) and (
          not pkg_resources.resource_isdir(module_base, module_file))
    except (ValueError, ImportError):
      return False

  @classmethod
  def matches(cls, loadable):
    return isinstance(loadable, Compatibility.stringy) and cls.resource_exists(loadable)

  @classmethod
  def compile_into(cls, context, from_path, config_file):
    actual_file = relativize(from_path, config_file)
    module_base, module_file = os.path.split(actual_file)
    module_base = module_base.replace(os.sep, '.')
    context.compile(from_path, config_file,
        pkg_resources.resource_string(module_base, module_file))


class LoadableMapExecutor(ConfigExecutor):
  @classmethod
  def matches(cls, loadable):
    return isinstance(loadable, dict)

  @classmethod
  def find_root_file(cls, loadable):
    for key in loadable.keys():
      from_file, config_file = ConfigContext.from_key(key)
      if from_file == cls.ROOT:
        return config_file

  @classmethod
  def from_filename(cls, stack):
    return reduce(relativize, stack, '')

  @classmethod
  def get(cls, loadable):
    deposit_stack = [cls.ROOT]

    def ast_executor(config_file, context):
      from_file = cls.from_filename(deposit_stack)
      deposit_stack.append(config_file)
      context.compile(from_file, config_file, loadable[ConfigContext.key(from_file, config_file)])
      deposit_stack.pop()

    return ast_executor, cls.find_root_file(loadable)


class FilelikeExecutor(ConfigExecutor):
  @classmethod
  def matches(cls, loadable):
    return hasattr(loadable, 'read') and callable(loadable.read)

  @classmethod
  def get(cls, loadable):
    def ast_executor(config, context):
      if config is not loadable:
        raise ValueError('You may not include() anything from filelike objects.')
      context.compile(cls.ROOT, '<resource: %s>' % loadable, loadable.read())
    return ast_executor, loadable


class Config(object):
  class Error(Exception): pass
  class InvalidConfigError(Error): pass
  class NotFound(Error): pass

  DEFAULT_SCHEMA = 'from pystachio import *'
  EXECUTORS = [
    FileExecutor,
    ResourceExecutor,
    FilelikeExecutor,
    LoadableMapExecutor
  ]
  ROOT = None

  @classmethod
  def choose_executor(cls, obj):
    for executor in cls.EXECUTORS:
      if executor.matches(obj):
        return executor.get(obj)
    raise cls.NotFound('Could not load resource %s' % obj)

  @classmethod
  def load_schema(cls, environment, schema=None):
    Compatibility.exec_function(
        compile(schema or cls.DEFAULT_SCHEMA, "<exec_function>", "exec"), environment)

  def __init__(self, loadable, schema=None):
    self._environment = {}
    self._loadables = {}
    self.load_schema(self._environment, schema)
    root_executor, initial_config = self.choose_executor(loadable)
    context = ConfigContext(self._environment, self._loadables)
    self._environment.update(include=lambda fn: root_executor(fn, context))
    try:
      root_executor(initial_config, context)
    except (SyntaxError, ValueError) as e:
      raise self.InvalidConfigError(str(e))

  @property
  def loadables(self):
    return self._loadables

  @property
  def environment(self):
    return self._environment



from .base import Object
from .compatibility import Compatibility
from .parsing import MustacheParser
from .typing import Type, TypeCheck, TypeFactory, TypeMetaclass


class SimpleObject(Object, Type):
  """
    A simply-valued (unnamable) object.
  """
  __slots__ = ('_value',)

  def __init__(self, value):
    self._value = value
    super(SimpleObject, self).__init__()

  def get(self):
    return self._value

  def dup(self):
    return self.__class__(self._value)

  def _my_cmp(self, other):
    if self.__class__ != other.__class__:
      return -1
    si, _ = self.interpolate()
    oi, _ = other.interpolate()
    if si._value < oi._value:
      return -1
    elif si._value > oi._value:
      return 1
    else:
      return 0

  def __hash__(self):
    return hash(self._value)

  def __eq__(self, other):
    return self._my_cmp(other) == 0

  def __lt__(self, other):
    return self._my_cmp(other) == -1

  def __gt__(self, other):
    return self._my_cmp(other) == 1

  def __le__(self, other):
    return self._my_cmp(other) <= 0

  def __ge__(self, other):
    return self._my_cmp(other) >= 0

  def __unicode__(self):
    si, _ = self.interpolate()
    return unicode(si._value)

  def __str__(self):
    si, _ = self.interpolate()
    return str(si._value)

  def __repr__(self):
    return '%s(%s)' % (self.__class__.__name__, str(self) if Compatibility.PY3 else unicode(self))

  def interpolate(self):
    if not isinstance(self._value, Compatibility.stringy):
      return self.__class__(self.coerce(self._value)), []
    else:
      joins, unbound = MustacheParser.resolve(self._value, *self.scopes())
      if unbound:
        return self.__class__(joins), unbound
      else:
        # XXX
        self_copy = self.copy()
        self_copy._value = self_copy.coerce(joins)
        return self_copy, unbound

  @classmethod
  def type_factory(cls):
    return cls.__name__

  @classmethod
  def type_parameters(cls):
    return ()


class String(SimpleObject):
  @classmethod
  def checker(cls, obj):
    assert isinstance(obj, String)
    if isinstance(obj._value, Compatibility.stringy):
      return TypeCheck.success()
    else:
      # TODO(wickman)  Perhaps we should mark uninterpolated Mustache objects as
      # intrinsically non-stringy, because String will never typecheck false given
      # its input constraints.
      return TypeCheck.failure("%s not a string" % repr(obj._value))

  @classmethod
  def coerce(cls, value):
    ACCEPTED_SOURCE_TYPES = Compatibility.stringy + Compatibility.numeric
    if not isinstance(value, ACCEPTED_SOURCE_TYPES):
      raise cls.CoercionError(value, cls)
    return str(value) if Compatibility.PY3 else unicode(value)


class StringFactory(TypeFactory):
  PROVIDES = 'String'
  @staticmethod
  def create(type_dict, *type_parameters):
    return String


class Integer(SimpleObject):
  @classmethod
  def checker(cls, obj):
    assert isinstance(obj, Integer)
    if isinstance(obj._value, Compatibility.integer):
      return TypeCheck.success()
    else:
      return TypeCheck.failure("%s not an integer" % repr(obj._value))

  @classmethod
  def coerce(cls, value):
    ACCEPTED_SOURCE_TYPES = Compatibility.numeric + Compatibility.stringy
    if not isinstance(value, ACCEPTED_SOURCE_TYPES):
      raise cls.CoercionError(value, cls)
    try:
      return int(value)
    except ValueError:
      raise cls.CoercionError(value, cls)


class IntegerFactory(TypeFactory):
  PROVIDES = 'Integer'
  @staticmethod
  def create(type_dict, *type_parameters):
    return Integer


class Float(SimpleObject):
  @classmethod
  def checker(cls, obj):
    assert isinstance(obj, Float)
    if isinstance(obj._value, Compatibility.real + Compatibility.integer):
      return TypeCheck.success()
    else:
      return TypeCheck.failure("%s not a float" % repr(obj._value))

  @classmethod
  def coerce(cls, value):
    ACCEPTED_SOURCE_TYPES = Compatibility.numeric + Compatibility.stringy
    if not isinstance(value, ACCEPTED_SOURCE_TYPES):
      raise cls.CoercionError(value, cls)
    try:
      return float(value)
    except ValueError:
      raise cls.CoercionError(value, cls)

class FloatFactory(TypeFactory):
  PROVIDES = 'Float'
  @staticmethod
  def create(type_dict, *type_parameters):
    return Float


class Boolean(SimpleObject):
  @classmethod
  def checker(cls, obj):
    assert isinstance(obj, Boolean)
    if isinstance(obj._value, bool):
      return TypeCheck.success()
    else:
      return TypeCheck.failure("%s not a boolean" % repr(obj._value))

  @classmethod
  def coerce(cls, value):
    ACCEPTED_SOURCE_TYPES = (bool,) + Compatibility.numeric + Compatibility.stringy
    if not isinstance(value, ACCEPTED_SOURCE_TYPES):
      raise cls.CoercionError(value, cls)

    if isinstance(value, bool):
      return value
    elif isinstance(value, Compatibility.stringy):
      if value.lower() in ("true", "1"):
        return True
      elif value.lower() in ("false", "0"):
        return False
      else:
        raise cls.CoercionError(value, cls)
    else:
      return bool(value)

class BooleanFactory(TypeFactory):
  PROVIDES = 'Boolean'
  @staticmethod
  def create(type_dict, *type_parameters):
    return Boolean


class EnumContainer(SimpleObject):
  def __init__(self, value):
    stringish = String(value)
    _, refs = stringish.interpolate()
    if not refs and value not in self.VALUES:
      raise ValueError('%s only accepts the following values: %s' % (
          self.__class__.__name__, ', '.join(self.VALUES)))
    super(EnumContainer, self).__init__(value)

  @classmethod
  def checker(cls, obj):
    assert isinstance(obj, EnumContainer)
    if isinstance(obj._value, Compatibility.stringy) and obj._value in cls.VALUES:
      return TypeCheck.success()
    else:
      return TypeCheck.failure("%s not in the enumeration (%s)" % (repr(obj._value),
        ', '.join(cls.VALUES)))

  @classmethod
  def coerce(cls, value):
    if not isinstance(value, Compatibility.stringy) or value not in cls.VALUES:
      raise cls.CoercionError(value, cls, '%s is not one of %s' % (
        value, ', '.join(cls.VALUES)))
    return str(value) if Compatibility.PY3 else unicode(value)

  @classmethod
  def type_factory(cls):
    return 'Enum'

  @classmethod
  def type_parameters(cls):
    return (cls.__name__, cls.VALUES)


class EnumFactory(TypeFactory):
  PROVIDES = 'Enum'

  @staticmethod
  def create(type_dict, *type_parameters):
    """
      EnumFactory.create(*type_parameters) expects:
        enumeration name, (enumeration values)
    """
    name, values = type_parameters
    assert isinstance(values, (list, tuple))
    for value in values:
      assert isinstance(value, Compatibility.stringy)
    return TypeMetaclass(str(name), (EnumContainer,), { 'VALUES': values })


def Enum(*stuff):
  # TODO(wickman) Check input
  if len(stuff) == 2 and isinstance(stuff[0], Compatibility.stringy) and (
      isinstance(stuff[1], (list, tuple))):
    name, values = stuff
    return TypeFactory.new({}, EnumFactory.PROVIDES, name, values)
  else:
    return TypeFactory.new({}, EnumFactory.PROVIDES, 'Enum_' + '_'.join(stuff), stuff)



import copy
import json
from collections import Mapping
from inspect import isclass

from .base import Environment, Object
from .naming import Namable, frozendict
from .typing import Type, TypeCheck, TypeFactory, TypeMetaclass


class Empty(object):
  """The Empty sentinel representing an unspecified field."""
  pass


class TypeSignature(object):
  """
    Type metadata for composite type schemas.
  """

  def __init__(self, cls, required=False, default=Empty):
    """Create an instance of a type signature.
    Args:
        cls (Class): the "type" of the object this signature represents.
        required (bool):
        default(object): an instance of the type for a default value. This
           should be either an instance of cls or something coercable to cls.
    """
    assert isclass(cls)
    assert issubclass(cls, Object)
    if default is not Empty and not isinstance(default, cls):
      self._default = cls(default)
    else:
      self._default = default
    self._cls = cls
    self._required = required

  def serialize(self):
    return (self.required,
            self.default.get() if not self.empty else (),
            self.empty,
            self.klazz.serialize_type())

  @staticmethod
  def deserialize(sig, type_dict):
    req, default, empty, klazz_schema = sig
    real_class = TypeFactory.new(type_dict, *klazz_schema)
    if not empty:
      return TypeSignature(real_class, default=real_class(default), required=req)
    else:
      return TypeSignature(real_class, required=req)

  def __hash__(self):
    return hash(
      self.klazz.serialize_type(),
      self.required,
      self.default,
      self.empty)

  def __eq__(self, other):
    return (self.klazz.serialize_type() == other.klazz.serialize_type() and
            self.required == other.required and
            self.default == other.default and
            self.empty == other.empty)

  def __ne__(self, other):
    return not (self == other)

  def __repr__(self):
    return 'TypeSignature(%s, required: %s, default: %s, empty: %s)' % (
      self.klazz.__name__, self.required, self.default, self.empty)

  @property
  def klazz(self):
    return self._cls

  @property
  def required(self):
    return self._required

  @property
  def default(self):
    return self._default

  @property
  def empty(self):
    return self._default is Empty

  @staticmethod
  def wrap(sig):
    """Convert a Python class into a type signature."""
    if isclass(sig) and issubclass(sig, Object):
      return TypeSignature(sig)
    elif isinstance(sig, TypeSignature):
      return sig


def Required(cls):
  """
    Helper to make composite types read succintly.  Wrap a type and make its
    specification required during type-checking of composite types.
  """
  return TypeSignature(cls, required=True)


def Default(cls, default):
  """
    Helper to make composite types read succintly.  Wrap a type and assign it a
    default if it is unspecified in the construction of the composite type.
  """
  return TypeSignature(cls, required=False, default=default)


class StructFactory(TypeFactory):
  PROVIDES = 'Struct'

  @staticmethod
  def create(type_dict, *type_parameters):
    """
      StructFactory.create(*type_parameters) expects:

        class name,
        ((binding requirement1,),
          (binding requirement2, bound_to_scope),
           ...),
        ((attribute_name1, attribute_sig1 (serialized)),
         (attribute_name2, attribute_sig2 ...),
         ...
         (attribute_nameN, ...))
    """
    name, parameters = type_parameters
    for param in parameters:
      assert isinstance(param, tuple)
    typemap = dict((attr, TypeSignature.deserialize(param, type_dict))
                   for attr, param in parameters)
    attributes = {'TYPEMAP': typemap}
    return TypeMetaclass(str(name), (Structural,), attributes)


class StructMetaclass(type):
  """
    Schema-extracting metaclass for Struct objects.
  """
  @staticmethod
  def attributes_to_parameters(attributes):
    parameters = []
    for attr_name, attr_value in attributes.items():
      sig = TypeSignature.wrap(attr_value)
      if sig:
        parameters.append((attr_name, sig.serialize()))
    return tuple(parameters)

  def __new__(mcs, name, parents, attributes):
    if any(parent.__name__ == 'Struct' for parent in parents):
      type_parameters = StructMetaclass.attributes_to_parameters(attributes)
      return TypeFactory.new({}, 'Struct', name, type_parameters)
    else:
      return type.__new__(mcs, name, parents, attributes)

class IsNotMappingError(ValueError):
  """Raised when a composite argument is not a Mapping"""

  def __init__(self, arg):
    self.arg = arg

  def __repr__(self):
    return 'IsNotMappingError({!r})'.format(self.arg)

StructMetaclassWrapper = StructMetaclass('StructMetaclassWrapper', (object,), {})
class Structural(Object, Type, Namable):
  """A Structural base type for composite objects."""
  __slots__ = ('_schema_data',)

  def __init__(self, *args, **kw):
    self._schema_data = frozendict((attr, value.default) for (attr, value) in self.TYPEMAP.items())
    for arg in args:
      if not isinstance(arg, Mapping):
        raise IsNotMappingError(arg)
      self._update_schema_data(**arg)
    self._update_schema_data(**copy.copy(kw))
    super(Structural, self).__init__()

  def get(self):
    return frozendict((k, v.get()) for k, v in self._schema_data.items() if v is not Empty)

  def _process_schema_attribute(self, attr, value):
    if attr not in self.TYPEMAP:
      raise AttributeError('Unknown schema attribute %s' % attr)
    schema_type = self.TYPEMAP[attr]
    if value is Empty:
      return Empty
    elif isinstance(value, schema_type.klazz):
      return value
    else:
      return schema_type.klazz(value)

  def _update_schema_data(self, **kw):
    for attr, value in kw.items():
      self._schema_data[attr] = self._process_schema_attribute(attr, value)

  def dup(self):
    return self.__class__(**self._schema_data)

  def __call__(self, **kw):
    new_self = self.copy()
    new_self._update_schema_data(**copy.copy(kw))
    return new_self

  def __hash__(self):
    return hash(self.get())

  def __eq__(self, other):
    if not isinstance(other, Structural): return False
    if self.TYPEMAP != other.TYPEMAP: return False
    si = self.interpolate()
    oi = other.interpolate()
    return si[0]._schema_data == oi[0]._schema_data

  def __repr__(self):
    si, _ = self.interpolate()
    return '%s(%s)' % (
      self.__class__.__name__,
      (',\n%s' % (' ' * (len(self.__class__.__name__) + 1))).join(
          '%s=%s' % (key, val) for key, val in si._schema_data.items() if val is not Empty)
    )

  def __getattr__(self, attr):
    if not hasattr(self, 'TYPEMAP'):
      raise AttributeError

    if attr.startswith('has_'):
      if attr[4:] in self.TYPEMAP:
        return lambda: self._schema_data[attr[4:]] != Empty

    if attr not in self.TYPEMAP:
      raise AttributeError("%s has no attribute %s" % (self.__class__.__name__, attr))

    return lambda: self.interpolate_key(attr)

  def check(self):
    scopes = self.scopes()
    for name, signature in self.TYPEMAP.items():
      if self._schema_data[name] is Empty and signature.required:
        return TypeCheck.failure('%s[%s] is required.' % (self.__class__.__name__, name))
      elif self._schema_data[name] is not Empty:
        type_check = self._schema_data[name].in_scope(*scopes).check()
        if type_check.ok():
          continue
        else:
          return TypeCheck.failure('%s[%s] failed: %s' % (self.__class__.__name__, name,
            type_check.message()))
    return TypeCheck.success()

  @classmethod
  def _cast_scopes_to_child(cls, scopes):
    return tuple(Environment({'super': scope}) for scope in scopes)

  def _self_scope(self):
    return Environment(dict((key, value) for (key, value) in self._schema_data.items()
                       if value is not Empty))

  def scopes(self):
    self_scope = self._self_scope()
    return (Environment({'self': self_scope}), self_scope,) + self._scopes + (
        self._cast_scopes_to_child(self._scopes))

  def interpolate(self):
    unbound = set()
    interpolated_schema_data = {}
    scopes = self.scopes()
    for key, value in self._schema_data.items():
      if value is Empty:
        interpolated_schema_data[key] = Empty
      else:
        vinterp, vunbound = value.in_scope(*scopes).interpolate()
        unbound.update(vunbound)
        interpolated_schema_data[key] = vinterp
    return self.__class__(**interpolated_schema_data), list(unbound)

  def interpolate_key(self, attribute):
    if self._schema_data[attribute] is Empty:
      return Empty
    vinterp, _ = self._schema_data[attribute].in_scope(*self.scopes()).interpolate()
    return self._process_schema_attribute(attribute, vinterp)

  @classmethod
  def type_factory(cls):
    return 'Struct'

  @classmethod
  def type_parameters(cls):
    attrs = []
    if hasattr(cls, 'TYPEMAP'):
      attrs = sorted([(attr, sig.serialize()) for attr, sig in cls.TYPEMAP.items()])
    return (cls.__name__, tuple(attrs))

  @classmethod
  def _filter_against_schema(cls, values):
    result = {}
    for key, val in values.items():
      if key not in cls.TYPEMAP:
        continue
      if issubclass(cls.TYPEMAP[key].klazz, Structural):
        result[key] = cls.TYPEMAP[key].klazz._filter_against_schema(val)
      else:
        result[key] = val
    return result

  @classmethod
  def json_load(cls, fp, strict=False):
    return cls(json.load(fp) if strict else cls._filter_against_schema(json.load(fp)))

  @classmethod
  def json_loads(cls, json_string, strict=False):
    return cls(json.loads(json_string) if strict
               else cls._filter_against_schema(json.loads(json_string)))

  def json_dump(self, fp):
    d, _ = self.interpolate()
    return json.dump(d.get(), fp)

  def json_dumps(self):
    d, _ = self.interpolate()
    return json.dumps(d.get())

  def find(self, ref):
    if not ref.is_dereference():
      raise Namable.NamingError(self, ref)
    name = ref.action().value
    if name not in self.TYPEMAP or self._schema_data[name] is Empty:
      raise Namable.NotFound(self, ref)
    else:
      namable = self._schema_data[name]
      if ref.rest().is_empty():
        return namable.in_scope(*self.scopes())
      else:
        if not isinstance(namable, Namable):
          raise Namable.Unnamable(namable)
        else:
          return namable.in_scope(*self.scopes()).find(ref.rest())


class Struct(StructMetaclassWrapper, Structural):
  """
    Schema-based composite objects, e.g.

      class Employee(Struct):
        first = Required(String)
        last  = Required(String)
        email = Required(String)
        phone = String

      Employee(first = "brian", last = "wickman", email = "wickman@twitter.com").check()

    They're purely functional data structures and behave more like functors.
    In other words they're immutable:

      >>> brian = Employee(first = "brian")
      >>> brian(last = "wickman")
      Employee(last=String(wickman), first=String(brian))
      >>> brian
      Employee(first=String(brian))
  """
  pass



__author__ = 'Brian Wickman'
__version__ = '0.8.3'
__license__ = 'MIT'


import sys

from .base import Environment
from .basic import Boolean, Enum, Float, Integer, String
from .choice import Choice
from .composite import Default, Empty, Required, Struct
from .container import List, Map
from .naming import Namable, Ref
from .parsing import MustacheParser
from .typing import Type, TypeCheck, TypeFactory

if sys.version_info < (2, 6, 5):
  raise ImportError("pystachio requires Python >= 2.6.5")
